from django import forms
from django.contrib.auth.forms import UserCreationForm
from django.core.exceptions import FieldDoesNotExist
from django.forms import ModelForm
from mixins.restrictions import RestrictedFormMixin

from .models import (  # AssetRestriction,
    Account,
    ActiveManagementPolicy,
    AssetGroup,
    AssetRule,
    Hook,
    NotificationSettings,
    OrderSettings,
    RiskModel,
    Signal,
    Strategy,
    Trade,
    TradingTime,
    User,
)

# flake8: noqa: E501


class NewUserForm(UserCreationForm):
    email = forms.EmailField(required=True)

    class Meta:
        model = User
        fields = (
            "username",
            "email",
            "first_name",
            "last_name",
            "password1",
            "password2",
        )

    def save(self, commit=True):
        user = super(NewUserForm, self).save(commit=False)
        user.email = self.cleaned_data["email"]
        if commit:
            user.save()
        return user


class CustomUserCreationForm(UserCreationForm):
    class Meta:
        model = User
        fields = "__all__"


class HookForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = Hook
        fields = (
            "name",
            "hook",
        )
        help_texts = {
            "name": "Name of the hook. Informational only.",
            "hook": "The URL slug to use for the hook. Make it unique.",
        }


class SignalForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = Signal
        fields = (
            "name",
            "signal",
            "hook",
            "type",
            "direction",
        )
        help_texts = {
            "name": "Name of the signal. Informational only.",
            "signal": "The name of the signal in Drakdoo. Copy it from there.",
            "hook": "The hook this signal belongs to.",
            "type": "Whether the signal is used for entering or exiting trades, or determining the trend.",
            "direction": "The direction of the signal. This is used to determine if the signal is a buy or sell.",
        }


class AccountForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = Account
        fields = (
            "name",
            "exchange",
            "api_key",
            "api_secret",
            "initial_balance",
            "sandbox",
            "enabled",
        )
        help_texts = {
            "name": "Name of the account. Informational only.",
            "exchange": "The exchange to use for this account.",
            "api_key": "The API key or username for the account.",
            "api_secret": "The API secret or password/token for the account.",
            "sandbox": "Whether to use the sandbox/demo or not.",
            "enabled": "Whether the account is enabled.",
            "initial_balance": "The initial balance of the account.",
        }


class StrategyForm(RestrictedFormMixin, ModelForm):
    fieldargs = {
        "entry_signals": {"type": "entry"},
        "exit_signals": {"type": "exit"},
        "trend_signals": {"type": "trend"},
    }

    # Filter for enabled accounts
    def __init__(self, *args, **kwargs):
        super(StrategyForm, self).__init__(*args, **kwargs)
        self.fields["account"].queryset = Account.objects.filter(enabled=True)

    class Meta:
        model = Strategy
        fields = (
            "name",
            "description",
            "account",
            "asset_group",
            "risk_model",
            "trading_times",
            "order_settings",
            "entry_signals",
            "exit_signals",
            "trend_signals",
            "signal_trading_enabled",
            "active_management_enabled",
            "active_management_policy",
            "enabled",
        )

        help_texts = {
            "name": "Name of the strategy. Informational only.",
            "description": "Description of the strategy. Informational only.",
            "account": "The account to use for this strategy.",
            "asset_group": "Asset groups determine which pairs can be traded.",
            "risk_model": "The risk model to use for this strategy. Highly recommended.",
            "trading_times": "When the strategy will place new trades.",
            "order_settings": "Order settings to use for this strategy.",
            "entry_signals": "Callbacks received to these signals will trigger a trade.",
            "exit_signals": "Callbacks received to these signals will close all trades for the symbol on the account.",
            "trend_signals": "Callbacks received to these signals will limit the trading direction of the given symbol to the callback direction until further notice.",
            "signal_trading_enabled": "Whether the strategy will place trades based on signals.",
            "active_management_enabled": "Whether the strategy will amend/remove trades on the account that violate the rules.",
            "active_management_policy": "The policy to use for active management.",
            "enabled": "Whether the strategy is enabled.",
        }

    entry_signals = forms.ModelMultipleChoiceField(
        queryset=Signal.objects.all(),
        widget=forms.CheckboxSelectMultiple,
        help_text=Meta.help_texts["entry_signals"],
        required=False,
    )
    exit_signals = forms.ModelMultipleChoiceField(
        queryset=Signal.objects.all(),
        widget=forms.CheckboxSelectMultiple,
        help_text=Meta.help_texts["exit_signals"],
        required=False,
    )
    trend_signals = forms.ModelMultipleChoiceField(
        queryset=Signal.objects.all(),
        widget=forms.CheckboxSelectMultiple,
        help_text=Meta.help_texts["trend_signals"],
        required=False,
    )
    trading_times = forms.ModelMultipleChoiceField(
        queryset=TradingTime.objects.all(), widget=forms.CheckboxSelectMultiple
    )

    def clean(self):
        cleaned_data = super(StrategyForm, self).clean()
        entry_signals = cleaned_data.get("entry_signals")
        exit_signals = cleaned_data.get("exit_signals")
        for entry in entry_signals.all():
            if entry in exit_signals.all():
                self._errors["entry_signals"] = self.error_class(
                    [
                        "You cannot bet against yourself. Do not use the same signal for entry and exit."
                    ]
                )
        for exit in exit_signals.all():
            if exit in entry_signals.all():
                self._errors["exit_signals"] = self.error_class(
                    [
                        "You cannot bet against yourself. Do not use the same signal for entry and exit."
                    ]
                )
        # Get all the directions for entry and exit signals
        entries_set = set([x.direction for x in entry_signals.all()])
        exits_set = set([x.direction for x in exit_signals.all()])

        # Make sure both fields are filled before we check this
        if entries_set and exits_set:
            # Combine them into one set
            check_set = set()
            check_set.update(entries_set, exits_set)

            # If the length is 1, they are all the same direction
            if len(check_set) == 1:
                self._errors["entry_signals"] = self.error_class(
                    [
                        "You cannot have entry and exit signals that are the same direction. At least one must be opposing."
                    ]
                )
                self._errors["exit_signals"] = self.error_class(
                    [
                        "You cannot have entry and exit signals that are the same direction. At least one must be opposing."
                    ]
                )
        if cleaned_data.get("active_management_enabled"):
            # Ensure that no other strategy with this account has active management enabled
            if (
                Strategy.objects.filter(
                    account=cleaned_data.get("account"),
                    active_management_enabled=True,
                    enabled=True,
                )
                .exclude(id=self.instance.id)
                .exists()
            ):
                self.add_error(
                    "active_management_enabled",
                    "You cannot have more than one strategy with active management enabled for the same account.",
                )
                return
            if not cleaned_data.get("active_management_policy"):
                self.add_error(
                    "active_management_policy",
                    "You must select an active management policy if active management is enabled.",
                )
                return
        return cleaned_data


class TradeForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = Trade
        fields = (
            "account",
            "symbol",
            "type",
            "time_in_force",
            "amount",
            "price",
            "stop_loss",
            "trailing_stop_loss",
            "take_profit",
            "direction",
        )
        help_texts = {
            "account": "The account to use for this trade.",
            "symbol": "The symbol to trade.",
            "type": "Market: Buy/Sell at the current market price. Limit: Buy/Sell at a specified price. Limits protect you more against market slippage.",
            "time_in_force": "The time in force controls how the order is executed.",
            "amount": "The amount to trade in the quote currency (the second part of the symbol if applicable, otherwise your account base currency).",
            "price": "The price to trade at. Sets this price as the trigger price for limit orders. Sets a price bound (if supported) for market orders.",
            "stop_loss": "The stop loss price. This will be set at the specified price.",
            "trailing_stop_loss": "The trailing stop loss price. This will be set at the specified price and will follow the price as it moves in your favor.",
            "take_profit": "The take profit price. This will be set at the specified price.",
            "direction": "The direction of the trade. This is used to determine if the trade is a buy or sell.",
        }


class TradingTimeForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = TradingTime
        fields = (
            "name",
            "description",
            "start_day",
            "start_time",
            "end_day",
            "end_time",
        )
        help_texts = {
            "name": "Name of the trading time. Informational only.",
            "description": "Description of the trading time. Informational only.",
            "start_day": "The day of the week to start trading.",
            "start_time": "The time of day to start trading.",
            "end_day": "The day of the week to stop trading.",
            "end_time": "The time of day to stop trading.",
        }


class NotificationSettingsForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = NotificationSettings
        fields = (
            "ntfy_topic",
            "ntfy_url",
        )
        help_texts = {
            "ntfy_topic": "The topic to send notifications to.",
            "ntfy_url": "Custom NTFY server. Leave blank to use the default server.",
        }


class RiskModelForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = RiskModel
        fields = (
            "name",
            "description",
            "max_loss_percent",
            "max_risk_percent",
            "max_open_trades",
            "max_open_trades_per_symbol",
            "price_slippage_percent",
            "callback_price_deviation_percent",
        )
        help_texts = {
            "name": "Name of the risk model. Informational only.",
            "description": "Description of the risk model. Informational only.",
            "max_loss_percent": "The maximum percent of the account balance that can be lost before we cease trading.",
            "max_risk_percent": "The maximum percent of the account balance that can be risked on all open trades.",
            "max_open_trades": "The maximum number of open trades.",
            "max_open_trades_per_symbol": "The maximum number of open trades per symbol.",
            "price_slippage_percent": "The price slippage is the maximum percent the price can move against you before the trade is cancelled. Limit orders will be set at this percentage against your favour. Market orders will have a price bound set if this is supported.",
            "callback_price_deviation_percent": "The callback price deviation is the maximum percent the price can change between receiving the callback and acting on it. This protects against rogue or delayed callbacks. Keep it low.",
        }


class AssetGroupForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = AssetGroup
        fields = (
            "name",
            "description",
            "when_no_data",
            "when_no_match",
            "when_no_aggregation",
            "when_not_in_bounds",
            "when_bullish",
            "when_bearish",
        )
        help_texts = {
            "name": "Name of the asset group. Informational only.",
            "description": "Description of the asset group. Informational only.",
            "when_no_data": "The action to take when no webhooks have been received for an asset.",
            "when_no_match": "The action to take when there were no matches last callback for an asset.",
            "when_no_aggregation": "The action to take when there is no defined aggregations for the asset.",
            "when_not_in_bounds": "The action to take when the aggregation is not breaching either bound.",
            "when_bullish": "The action to take when the asset is bullish.",
            "when_bearish": "The action to take when the asset is bearish.",
        }


class AssetRuleForm(RestrictedFormMixin, ModelForm):
    def __init__(self, *args, **kwargs):
        super(AssetRuleForm, self).__init__(*args, **kwargs)
        self.fields["value"].disabled = True
        self.fields["original_status"].disabled = True
        self.fields["aggregation"].disabled = True

    class Meta:
        model = AssetRule
        fields = (
            "asset",
            "aggregation",
            "value",
            "original_status",
            "status",
            "trigger_below",
            "trigger_above",
        )

        help_texts = {
            "asset": "The asset to apply the rule to.",
            "aggregation": "Aggregation of the callback",
            "value": "Value of the aggregation",
            "original_status": "The original status of the asset.",
            "status": "The status of the asset, following rules configured on the Asset Group.",
            "trigger_below": "Trigger Bearish when value is below this.",
            "trigger_above": "Trigger Bullish when value is above this.",
        }


class OrderSettingsForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = OrderSettings
        fields = (
            "name",
            "description",
            "order_type",
            "time_in_force",
            "take_profit_percent",
            "stop_loss_percent",
            "trailing_stop_loss_percent",
            "trade_size_percent",
        )
        help_texts = {
            "name": "Name of the order settings. Informational only.",
            "description": "Description of the order settings. Informational only.",
            "order_type": "Market: Buy/Sell at the current market price. Limit: Buy/Sell at a specified price. Limits protect you more against market slippage.",
            "time_in_force": "The time in force controls how the order is executed.",
            "take_profit_percent": "The take profit will be set at this percentage above/below the entry price.",
            "stop_loss_percent": "The stop loss will be set at this percentage above/below the entry price.",
            "trailing_stop_loss_percent": "The trailing stop loss will be set at this percentage above/below the entry price. A trailing stop loss will follow the price as it moves in your favor.",
            "trade_size_percent": "Percentage of the account balance to use for each trade.",
        }


class ActiveManagementPolicyForm(RestrictedFormMixin, ModelForm):
    class Meta:
        model = ActiveManagementPolicy
        fields = (
            "name",
            "description",
            "when_trading_time_violated",
            "when_trends_violated",
            "when_position_size_violated",
            "when_protection_violated",
            "when_asset_groups_violated",
            "when_max_open_trades_violated",
            "when_max_open_trades_per_symbol_violated",
            "when_max_loss_violated",
            "when_max_risk_violated",
            "when_crossfilter_violated",
        )
        help_texts = {
            "name": "Name of the active management policy. Informational only.",
            "description": "Description of the active management policy. Informational only.",
            "when_trading_time_violated": "The action to take when the trading time is violated.",
            "when_trends_violated": "The action to take a trade against the trend is discovered.",
            "when_position_size_violated": "The action to take when a trade exceeding the position size is discovered.",
            "when_protection_violated": "The action to take when a trade violating/lacking defined TP/SL/TSL is discovered.",
            "when_asset_groups_violated": "The action to take when a trade violating the asset group rules is discovered.",
            "when_max_open_trades_violated": "The action to take when a trade puts the account above the maximum open trades.",
            "when_max_open_trades_per_symbol_violated": "The action to take when a trade puts the account above the maximum open trades per symbol.",
            "when_max_loss_violated": "The action to take when the account exceeds its maximum loss. NOTE: The close action will close all trades.",
            "when_max_risk_violated": "The action to take when a trade exposes the account to more than the maximum risk.",
            "when_crossfilter_violated": "The action to take when a trade is deemed to conflict with another -- e.g. a buy and sell on the same asset.",
        }
